@@ -1,7 +1,7 @@
 /*
  * server.c - Provide shadowsocks service
  *
- * Copyright (C) 2013 - 2016, Max Lv <max.c.lv@gmail.com>
+ * Copyright (C) 2013 - 2017, Max Lv <max.c.lv@gmail.com>
  *
  * This file is part of the shadowsocks-libev.
  *
@@ -75,7 +75,19 @@
 #endif
 
 #ifndef MAX_FRAG
-#define MAX_FRAG 2
+#define MAX_FRAG 1
+#endif
+
+#ifdef USE_NFCONNTRACK_TOS
+
+#ifndef MARK_MAX_PACKET
+#define MARK_MAX_PACKET 10
+#endif
+
+#ifndef MARK_MASK_PREFIX
+#define MARK_MASK_PREFIX 0xDC00
+#endif
+
 #endif
 
 static void signal_cb(EV_P_ ev_signal *w, int revents);
@@ -99,14 +111,13 @@ static void close_and_free_server(EV_P_ server_t *server);
 static void server_resolve_cb(struct sockaddr *addr, void *data);
 static void query_free_cb(void *data);
 
-static size_t parse_header_len(const char atyp, const char *data, size_t offset);
-static int is_header_complete(const buffer_t *buf);
+int verbose     = 0;
+int reuse_port = 0;
 
-int verbose = 0;
+static crypto_t *crypto;
 
 static int acl       = 0;
 static int mode      = TCP_ONLY;
-static int auth      = 0;
 static int ipv6first = 0;
 static int fast_open = 0;
 
@@ -118,7 +129,7 @@ static int server_conn = 0;
 
 static char *plugin          = NULL;
 static char *bind_address    = NULL;
-static char *server_port     = NULL;
+static char *remote_port     = NULL;
 static char *manager_address = NULL;
 uint64_t tx                  = 0;
 uint64_t rx                  = 0;
@@ -143,7 +154,7 @@ stat_update_cb(EV_P_ ev_timer *watcher, int revents)
         LOGI("update traffic stat: tx: %" PRIu64 " rx: %" PRIu64 "", tx, rx);
     }
 
-    snprintf(resp, BUF_SIZE, "stat: {\"%s\":%" PRIu64 "}", server_port, tx + rx);
+    snprintf(resp, BUF_SIZE, "stat: {\"%s\":%" PRIu64 "}", remote_port, tx + rx);
     msgLen = strlen(resp) + 1;
 
     ss_addr_t ip_addr = { .host = NULL, .port = NULL };
@@ -158,7 +169,7 @@ stat_update_cb(EV_P_ ev_timer *watcher, int revents)
 
         memset(&claddr, 0, sizeof(struct sockaddr_un));
         claddr.sun_family = AF_UNIX;
-        snprintf(claddr.sun_path, sizeof(claddr.sun_path), "/tmp/shadowsocks.%s", server_port);
+        snprintf(claddr.sun_path, sizeof(claddr.sun_path), "/tmp/shadowsocks.%s", remote_port);
 
         unlink(claddr.sun_path);
 
@@ -219,66 +230,6 @@ free_connections(struct ev_loop *loop)
     }
 }
 
-static size_t
-parse_header_len(const char atyp, const char *data, size_t offset)
-{
-    size_t len = 0;
-    if ((atyp & ADDRTYPE_MASK) == 1) {
-        // IP V4
-        len += sizeof(struct in_addr);
-    } else if ((atyp & ADDRTYPE_MASK) == 3) {
-        // Domain name
-        uint8_t name_len = *(uint8_t *)(data + offset);
-        len += name_len + 1;
-    } else if ((atyp & ADDRTYPE_MASK) == 4) {
-        // IP V6
-        len += sizeof(struct in6_addr);
-    } else {
-        return 0;
-    }
-    len += 2;
-    return len;
-}
-
-static int
-is_header_complete(const buffer_t *buf)
-{
-    size_t header_len = 0;
-    size_t buf_len    = buf->len;
-
-    char atyp = buf->data[header_len];
-
-    // 1 byte for atyp
-    header_len++;
-
-    if ((atyp & ADDRTYPE_MASK) == 1) {
-        // IP V4
-        header_len += sizeof(struct in_addr);
-    } else if ((atyp & ADDRTYPE_MASK) == 3) {
-        // Domain name
-        // domain len + len of domain
-        if (buf_len < header_len + 1)
-            return 0;
-        uint8_t name_len = *(uint8_t *)(buf->data + header_len);
-        header_len += name_len + 1;
-    } else if ((atyp & ADDRTYPE_MASK) == 4) {
-        // IP V6
-        header_len += sizeof(struct in6_addr);
-    } else {
-        return -1;
-    }
-
-    // len of port
-    header_len += 2;
-
-    // size of ONETIMEAUTH_BYTES
-    if (auth || (atyp & ONETIMEAUTH_FLAG)) {
-        header_len += ONETIMEAUTH_BYTES;
-    }
-
-    return buf_len >= header_len ? 1 : 0;
-}
-
 static char *
 get_peer_name(int fd)
 {
@@ -326,7 +277,7 @@ reset_addr(int fd)
 }
 
 static void
-report_addr(int fd, int err_level)
+report_addr(int fd, int err_level, const char* info)
 {
 #ifdef __linux__
     set_linger(fd);
@@ -335,7 +286,7 @@ report_addr(int fd, int err_level)
     char *peer_name;
     peer_name = get_peer_name(fd);
     if (peer_name != NULL) {
-        LOGE("failed to handshake with %s", peer_name);
+        LOGE("failed to handshake with %s: %s", peer_name, info);
         update_block_list(peer_name, err_level);
     }
 }
@@ -389,6 +340,8 @@ create_and_bind(const char *host, const char *port, int mptcp)
     hints.ai_flags    = AI_PASSIVE | AI_ADDRCONFIG; /* For wildcard IP address */
     hints.ai_protocol = IPPROTO_TCP;
 
+    result = NULL;
+
     for (int i = 1; i < 8; i++) {
         s = getaddrinfo(host, port, &hints, &result);
         if (s == 0) {
@@ -404,6 +357,11 @@ create_and_bind(const char *host, const char *port, int mptcp)
         return -1;
     }
 
+    if (result == NULL) {
+        LOGE("Could not bind");
+        return -1;
+    }
+
     rp = result;
 
     /*
@@ -441,14 +399,23 @@ create_and_bind(const char *host, const char *port, int mptcp)
 #ifdef SO_NOSIGPIPE
         setsockopt(listen_sock, SOL_SOCKET, SO_NOSIGPIPE, &opt, sizeof(opt));
 #endif
-        int err = set_reuseport(listen_sock);
-        if (err == 0) {
-            LOGI("tcp port reuse enabled");
+        if (reuse_port) {
+            int err = set_reuseport(listen_sock);
+            if (err == 0) {
+                LOGI("tcp port reuse enabled");
+            }
         }
 
         if (mptcp == 1) {
-            int err = setsockopt(listen_sock, SOL_TCP, MPTCP_ENABLED, &opt, sizeof(opt));
-            if (err == -1) {
+            int i = 0;
+            while((mptcp = mptcp_enabled_values[i]) > 0) {
+                int err = setsockopt(listen_sock, IPPROTO_TCP, mptcp, &opt, sizeof(opt));
+                if (err != -1) {
+                    break;
+                }
+                i++;
+            }
+            if (mptcp == 0) {
                 ERROR("failed to enable multipath TCP");
             }
         }
@@ -462,11 +429,7 @@ create_and_bind(const char *host, const char *port, int mptcp)
         }
 
         close(listen_sock);
-    }
-
-    if (rp == NULL) {
-        LOGE("Could not bind");
-        return -1;
+        listen_sock = -1;
     }
 
     freeaddrinfo(result);
@@ -600,6 +563,93 @@ connect_to_remote(EV_P_ struct addrinfo *res,
     return remote;
 }
 
+#ifdef USE_NFCONNTRACK_TOS
+int setMarkDscpCallback(enum nf_conntrack_msg_type type, struct nf_conntrack *ct, void *data)
+{
+	server_t* server = (server_t*) data;
+	struct dscptracker* tracker = server->tracker;
+
+	tracker->mark = nfct_get_attr_u32(ct, ATTR_MARK);
+	if ((tracker->mark & 0xff00) == MARK_MASK_PREFIX) {
+		// Extract DSCP value from mark value
+		tracker->dscp = tracker->mark & 0x00ff;
+		int tos = (tracker->dscp) << 2;
+		if (setsockopt(server->fd, IPPROTO_IP, IP_TOS, &tos, sizeof(tos)) != 0) {
+			ERROR("iptable setsockopt IP_TOS");
+		};
+	}
+	return NFCT_CB_CONTINUE;
+}
+
+void conntrackQuery(server_t* server) {
+	struct dscptracker* tracker = server->tracker;
+	if(tracker && tracker->ct) {
+		// Trying query mark from nf conntrack
+		struct nfct_handle *h = nfct_open(CONNTRACK, 0);
+		if (h) {
+			nfct_callback_register(h, NFCT_T_ALL, setMarkDscpCallback, (void*) server);
+			int x = nfct_query(h, NFCT_Q_GET, tracker->ct);
+			if (x == -1) {
+				LOGE("QOS: Failed to retrieve connection mark %s", strerror(errno));
+			}
+			nfct_close(h);
+		} else {
+			LOGE("QOS: Failed to open conntrack handle for upstream netfilter mark retrieval.");
+		}
+	}
+}
+
+void setTosFromConnmark(remote_t* remote, server_t* server)
+{
+	if(server->tracker && server->tracker->ct) {
+		if(server->tracker->mark == 0 && server->tracker->packet_count < MARK_MAX_PACKET) {
+			server->tracker->packet_count++;
+			conntrackQuery(server);
+		}
+	} else {
+		socklen_t len;
+		struct sockaddr_storage sin;
+		len = sizeof(sin);
+		if (getsockname(remote->fd, (struct sockaddr *)&sin, &len) == 0) {
+			struct sockaddr_storage from_addr;
+			len = sizeof from_addr;
+			if(getpeername(remote->fd, (struct sockaddr*)&from_addr, &len) == 0) {
+				if((server->tracker = (struct dscptracker*) malloc(sizeof(struct dscptracker))))
+				{
+					if ((server->tracker->ct = nfct_new())) {
+						// Build conntrack query SELECT
+						if (from_addr.ss_family == AF_INET) {
+							struct sockaddr_in *src = (struct sockaddr_in *)&from_addr;
+							struct sockaddr_in *dst = (struct sockaddr_in *)&sin;
+
+							nfct_set_attr_u8(server->tracker->ct, ATTR_L3PROTO, AF_INET);
+							nfct_set_attr_u32(server->tracker->ct, ATTR_IPV4_DST, dst->sin_addr.s_addr);
+							nfct_set_attr_u32(server->tracker->ct, ATTR_IPV4_SRC, src->sin_addr.s_addr);
+							nfct_set_attr_u16(server->tracker->ct, ATTR_PORT_DST, dst->sin_port);
+							nfct_set_attr_u16(server->tracker->ct, ATTR_PORT_SRC, src->sin_port);
+						} else if (from_addr.ss_family == AF_INET6) {
+							struct sockaddr_in6 *src = (struct sockaddr_in6 *)&from_addr;
+							struct sockaddr_in6 *dst = (struct sockaddr_in6 *)&sin;
+
+							nfct_set_attr_u8(server->tracker->ct, ATTR_L3PROTO, AF_INET6);
+							nfct_set_attr(server->tracker->ct, ATTR_IPV6_DST, dst->sin6_addr.s6_addr);
+							nfct_set_attr(server->tracker->ct, ATTR_IPV6_SRC, src->sin6_addr.s6_addr);
+							nfct_set_attr_u16(server->tracker->ct, ATTR_PORT_DST, dst->sin6_port);
+							nfct_set_attr_u16(server->tracker->ct, ATTR_PORT_SRC, src->sin6_port);
+						}
+						nfct_set_attr_u8(server->tracker->ct, ATTR_L4PROTO, IPPROTO_TCP);
+						conntrackQuery(server);
+					} else {
+						LOGE("Failed to allocate new conntrack for upstream netfilter mark retrieval.");
+						server->tracker->ct=NULL;
+					};
+				}
+			}
+		}
+	}
+}
+#endif
+
 static void
 server_recv_cb(EV_P_ ev_io *w, int revents)
 {
@@ -607,25 +657,16 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
     server_t *server              = server_recv_ctx->server;
     remote_t *remote              = NULL;
 
-    int len       = server->buf->len;
     buffer_t *buf = server->buf;
 
-    if (server->stage > STAGE_PARSE) {
+    if (server->stage == STAGE_STREAM) {
         remote = server->remote;
         buf    = remote->buf;
-        len    = 0;
 
         ev_timer_again(EV_A_ & server->recv_ctx->watcher);
     }
 
-    if (len > BUF_SIZE) {
-        ERROR("out of recv buffer");
-        close_and_free_remote(EV_A_ remote);
-        close_and_free_server(EV_A_ server);
-        return;
-    }
-
-    ssize_t r = recv(server->fd, buf->data + len, BUF_SIZE - len, 0);
+    ssize_t r = recv(server->fd, buf->data, BUF_SIZE, 0);
 
     if (r == 0) {
         // connection closed
@@ -649,95 +690,27 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
     }
 
     tx += r;
+    buf->len = r;
 
-    if (server->frag >= MAX_FRAG) {
-        LOGE("fragmentation detected");
+    int err = crypto->decrypt(buf, server->d_ctx, BUF_SIZE);
+
+    if (err == CRYPTO_ERROR) {
+        report_addr(server->fd, MALICIOUS, "authentication error");
         close_and_free_remote(EV_A_ remote);
         close_and_free_server(EV_A_ server);
         return;
-    }
-
-    if (server->stage == STAGE_ERROR) {
-        server->buf->len = 0;
-        server->buf->idx = 0;
+    } else if (err == CRYPTO_NEED_MORE) {
+        if (server->stage != STAGE_STREAM && server->frag > MAX_FRAG) {
+            report_addr(server->fd, MALICIOUS, "malicious fragmentation");
+            close_and_free_remote(EV_A_ remote);
+            close_and_free_server(EV_A_ server);
+        }
+        server->frag++;
         return;
     }
 
-    // handle incomplete header part 1
-    if (server->stage == STAGE_INIT) {
-        buf->len += r;
-
-        if (buf->len <= enc_get_iv_len() + 1) {
-            // wait for more
-            server->frag++;
-            return;
-        }
-    } else {
-        buf->len = r;
-    }
-
-    int err = ss_decrypt(buf, server->d_ctx, BUF_SIZE);
-
-    if (err) {
-        report_addr(server->fd, MALICIOUS);
-        close_and_free_remote(EV_A_ remote);
-        close_and_free_server(EV_A_ server);
-        return;
-    }
-
-    // handle incomplete header part 2
-    if (server->stage == STAGE_INIT) {
-        int ret = is_header_complete(server->buf);
-        if (ret == 1) {
-            bfree(server->header_buf);
-            ss_free(server->header_buf);
-            server->stage = STAGE_PARSE;
-        } else if (ret == -1) {
-            server->stage = STAGE_ERROR;
-            report_addr(server->fd, MALFORMED);
-            server->buf->len = 0;
-            server->buf->idx = 0;
-            return;
-        } else {
-            server->stage = STAGE_HANDSHAKE;
-        }
-    }
-
-    if (server->stage == STAGE_HANDSHAKE) {
-        size_t header_len = server->header_buf->len;
-        brealloc(server->header_buf, server->buf->len + header_len, BUF_SIZE);
-        memcpy(server->header_buf->data + header_len,
-               server->buf->data, server->buf->len);
-        server->header_buf->len = server->buf->len + header_len;
-
-        int ret = is_header_complete(server->buf);
-
-        if (ret == 1) {
-            brealloc(server->buf, server->header_buf->len, BUF_SIZE);
-            memcpy(server->buf->data, server->header_buf->data, server->header_buf->len);
-            server->buf->len = server->header_buf->len;
-            bfree(server->header_buf);
-            ss_free(server->header_buf);
-            server->stage = STAGE_PARSE;
-        } else {
-            if (ret == -1)
-                server->stage = STAGE_ERROR;
-            server->frag++;
-            server->buf->len = 0;
-            server->buf->idx = 0;
-            return;
-        }
-    }
-
     // handshake and transmit data
     if (server->stage == STAGE_STREAM) {
-        if (server->auth && !ss_check_hash(remote->buf, server->chunk, server->d_ctx, BUF_SIZE)) {
-            LOGE("hash error");
-            report_addr(server->fd, BAD);
-            close_and_free_server(EV_A_ server);
-            close_and_free_remote(EV_A_ remote);
-            return;
-        }
         int s = send(remote->fd, remote->buf->data, remote->buf->len, 0);
         if (s == -1) {
             if (errno == EAGAIN || errno == EWOULDBLOCK) {
@@ -757,7 +730,7 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
             ev_io_start(EV_A_ & remote->send_ctx->io);
         }
         return;
-    } else if (server->stage == STAGE_PARSE) {
+    } else if (server->stage == STAGE_INIT) {
         /*
          * Shadowsocks TCP Relay Header:
          *
@@ -767,24 +740,6 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
          *    |  1   | Variable |    2     |      10        |
          *    +------+----------+----------+----------------+
          *
-         *    If ATYP & ONETIMEAUTH_FLAG(0x10) != 0, Authentication (HMAC-SHA1) is enabled.
-         *
-         *    The key of HMAC-SHA1 is (IV + KEY) and the input is the whole header.
-         *    The output of HMAC-SHA is truncated to 10 bytes (leftmost bits).
-         */
-
-        /*
-         * Shadowsocks Request's Chunk Authentication for TCP Relay's payload
-         * (No chunk authentication for response's payload):
-         *
-         *    +------+-----------+-------------+------+
-         *    | LEN  | HMAC-SHA1 |    DATA     |      ...
-         *    +------+-----------+-------------+------+
-         *    |  2   |    10     |  Variable   |      ...
-         *    +------+-----------+-------------+------+
-         *
-         *    The key of HMAC-SHA1 is (IV + CHUNK ID)
-         *    The output of HMAC-SHA is truncated to 10 bytes (leftmost bits).
          */
 
         int offset     = 0;
@@ -797,27 +752,6 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
         memset(&info, 0, sizeof(struct addrinfo));
         memset(&storage, 0, sizeof(struct sockaddr_storage));
 
-        if (auth || (atyp & ONETIMEAUTH_FLAG)) {
-            size_t header_len = parse_header_len(atyp, server->buf->data, offset);
-            size_t len        = server->buf->len;
-
-            if (header_len == 0 || len < offset + header_len + ONETIMEAUTH_BYTES) {
-                report_addr(server->fd, MALFORMED);
-                close_and_free_server(EV_A_ server);
-                return;
-            }
-
-            server->buf->len = offset + header_len + ONETIMEAUTH_BYTES;
-            if (ss_onetimeauth_verify(server->buf, server->d_ctx->evp.iv)) {
-                report_addr(server->fd, BAD);
-                close_and_free_server(EV_A_ server);
-                return;
-            }
-
-            server->buf->len = len;
-            server->auth     = 1;
-        }
-
         // get remote addr and port
         if ((atyp & ADDRTYPE_MASK) == 1) {
             // IP V4
@@ -830,8 +764,7 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
                          host, INET_ADDRSTRLEN);
                 offset += in_addr_len;
             } else {
-                LOGE("invalid header with addr type %d", atyp);
-                report_addr(server->fd, MALFORMED);
+                report_addr(server->fd, MALFORMED, "invalid length for ipv4 address");
                 close_and_free_server(EV_A_ server);
                 return;
             }
@@ -848,8 +781,7 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
                 memcpy(host, server->buf->data + offset + 1, name_len);
                 offset += name_len + 1;
             } else {
-                LOGE("invalid name length: %d", name_len);
-                report_addr(server->fd, MALFORMED);
+                report_addr(server->fd, MALFORMED, "invalid host name length");
                 close_and_free_server(EV_A_ server);
                 return;
             }
@@ -882,8 +814,7 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
                 }
             } else {
                 if (!validate_hostname(host, name_len)) {
-                    LOGE("invalid host name");
-                    report_addr(server->fd, MALFORMED);
+                    report_addr(server->fd, MALFORMED, "invalid host name");
                     close_and_free_server(EV_A_ server);
                     return;
                 }
@@ -901,7 +832,7 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
                 offset += in6_addr_len;
             } else {
                 LOGE("invalid header with addr type %d", atyp);
-                report_addr(server->fd, MALFORMED);
+                report_addr(server->fd, MALFORMED, "invalid length for ipv6 address");
                 close_and_free_server(EV_A_ server);
                 return;
             }
@@ -914,8 +845,7 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
         }
 
         if (offset == 1) {
-            LOGE("invalid header with addr type %d", atyp);
-            report_addr(server->fd, MALFORMED);
+            report_addr(server->fd, MALFORMED, "invalid address type");
             close_and_free_server(EV_A_ server);
             return;
         }
@@ -924,12 +854,8 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
 
         offset += 2;
 
-        if (server->auth) {
-            offset += ONETIMEAUTH_BYTES;
-        }
-
         if (server->buf->len < offset) {
-            report_addr(server->fd, MALFORMED);
+            report_addr(server->fd, MALFORMED, "invalid request length");
             close_and_free_server(EV_A_ server);
             return;
         } else {
@@ -944,13 +870,6 @@ server_recv_cb(EV_P_ ev_io *w, int revents)
                 LOGI("connect to %s:%d", host, ntohs(port));
         }
 
-        if (server->auth && !ss_check_hash(server->buf, server->chunk, server->d_ctx, BUF_SIZE)) {
-            LOGE("hash error");
-            report_addr(server->fd, BAD);
-            close_and_free_server(EV_A_ server);
-            return;
-        }
-
         if (!need_query) {
             remote_t *remote = connect_to_remote(EV_A_ & info, server);
 
@@ -1067,15 +986,6 @@ server_timeout_cb(EV_P_ ev_timer *watcher, int revents)
         LOGI("TCP connection timeout");
     }
 
-    if (server->stage < STAGE_PARSE) {
-        if (verbose) {
-            size_t len = server->stage ?
-                         server->header_buf->len : server->buf->len;
-            LOGI("incomplete header: %zu", len);
-        }
-        report_addr(server->fd, SUSPICIOUS);
-    }
-
     close_and_free_remote(EV_A_ remote);
     close_and_free_server(EV_A_ server);
 }
@@ -1184,7 +1094,7 @@ remote_recv_cb(EV_P_ ev_io *w, int revents)
     rx += r;
 
     server->buf->len = r;
-    int err = ss_encrypt(server->buf, server->e_ctx, BUF_SIZE);
+    int err = crypto->encrypt(server->buf, server->e_ctx, BUF_SIZE);
 
     if (err) {
         LOGE("invalid password or cipher");
@@ -1193,6 +1103,9 @@ remote_recv_cb(EV_P_ ev_io *w, int revents)
         return;
     }
 
+#ifdef USE_NFCONNTRACK_TOS
+    setTosFromConnmark(remote, server);
+#endif
     int s = send(server->fd, server->buf->data, server->buf->len, 0);
 
     if (s == -1) {
@@ -1386,11 +1299,9 @@ new_server(int fd, listen_ctx_t *listener)
     server->recv_ctx   = ss_malloc(sizeof(server_ctx_t));
     server->send_ctx   = ss_malloc(sizeof(server_ctx_t));
     server->buf        = ss_malloc(sizeof(buffer_t));
-    server->header_buf = ss_malloc(sizeof(buffer_t));
     memset(server->recv_ctx, 0, sizeof(server_ctx_t));
     memset(server->send_ctx, 0, sizeof(server_ctx_t));
     balloc(server->buf, BUF_SIZE);
-    balloc(server->header_buf, BUF_SIZE);
     server->fd                  = fd;
     server->recv_ctx->server    = server;
     server->recv_ctx->connected = 0;
@@ -1402,15 +1313,10 @@ new_server(int fd, listen_ctx_t *listener)
     server->listen_ctx          = listener;
     server->remote              = NULL;
 
-    if (listener->method) {
-        server->e_ctx = ss_malloc(sizeof(enc_ctx_t));
-        server->d_ctx = ss_malloc(sizeof(enc_ctx_t));
-        enc_ctx_init(listener->method, server->e_ctx, 1);
-        enc_ctx_init(listener->method, server->d_ctx, 0);
-    } else {
-        server->e_ctx = NULL;
-        server->d_ctx = NULL;
-    }
+    server->e_ctx = ss_align(sizeof(cipher_ctx_t));
+    server->d_ctx = ss_align(sizeof(cipher_ctx_t));
+    crypto->ctx_init(crypto->cipher, server->e_ctx, 1);
+    crypto->ctx_init(crypto->cipher, server->d_ctx, 0);
 
     int request_timeout = min(MAX_REQUEST_TIMEOUT, listener->timeout)
                           + rand() % MAX_REQUEST_TIMEOUT;
@@ -1420,11 +1326,6 @@ new_server(int fd, listen_ctx_t *listener)
     ev_timer_init(&server->recv_ctx->watcher, server_timeout_cb,
                   request_timeout, listener->timeout);
 
-    server->chunk = ss_malloc(sizeof(chunk_t));
-    memset(server->chunk, 0, sizeof(chunk_t));
-    server->chunk->buf = ss_malloc(sizeof(buffer_t));
-    memset(server->chunk->buf, 0, sizeof(buffer_t));
-
     cork_dllist_add(&connections, &server->entries);
 
     return server;
@@ -1433,34 +1334,34 @@ new_server(int fd, listen_ctx_t *listener)
 static void
 free_server(server_t *server)
 {
+#ifdef USE_NFCONNTRACK_TOS
+    if(server->tracker) {
+        struct dscptracker* tracker = server->tracker;
+        struct nf_conntrack* ct = server->tracker->ct;
+        server->tracker = NULL;
+        if (ct) {
+            nfct_destroy(ct);
+       }
+       free(tracker);
+    };
+#endif
     cork_dllist_remove(&server->entries);
 
-    if (server->chunk != NULL) {
-        if (server->chunk->buf != NULL) {
-            bfree(server->chunk->buf);
-            ss_free(server->chunk->buf);
-        }
-        ss_free(server->chunk);
-    }
     if (server->remote != NULL) {
         server->remote->server = NULL;
     }
     if (server->e_ctx != NULL) {
-        cipher_context_release(&server->e_ctx->evp);
+        crypto->ctx_release(server->e_ctx);
         ss_free(server->e_ctx);
     }
     if (server->d_ctx != NULL) {
-        cipher_context_release(&server->d_ctx->evp);
+        crypto->ctx_release(server->d_ctx);
         ss_free(server->d_ctx);
     }
     if (server->buf != NULL) {
         bfree(server->buf);
         ss_free(server->buf);
     }
-    if (server->header_buf != NULL) {
-        bfree(server->header_buf);
-        ss_free(server->header_buf);
-    }
 
     ss_free(server->recv_ctx);
     ss_free(server->send_ctx);
@@ -1566,12 +1467,14 @@ main(int argc, char **argv)
     int mtu         = 0;
     char *user      = NULL;
     char *password  = NULL;
+    char *key       = NULL;
     char *timeout   = NULL;
     char *method    = NULL;
     char *pid_path  = NULL;
     char *conf_path = NULL;
     char *iface     = NULL;
 
+    char *server_port = NULL;
     char *plugin_opts = NULL;
     char *plugin_port = NULL;
     char tmp_port[8];
@@ -1582,50 +1485,60 @@ main(int argc, char **argv)
     char *nameservers[MAX_DNS_NUM + 1];
     int nameserver_num = 0;
 
-    int option_index                    = 0;
     static struct option long_options[] = {
-        { "fast-open",       no_argument,       0, 0 },
-        { "acl",             required_argument, 0, 0 },
-        { "manager-address", required_argument, 0, 0 },
-        { "mtu",             required_argument, 0, 0 },
-        { "help",            no_argument,       0, 0 },
-        { "plugin",          required_argument, 0, 0 },
-        { "plugin-opts",     required_argument, 0, 0 },
+        { "fast-open",       no_argument,       NULL, GETOPT_VAL_FAST_OPEN },
+        { "reuse-port",      no_argument,       NULL, GETOPT_VAL_REUSE_PORT },
+        { "acl",             required_argument, NULL, GETOPT_VAL_ACL },
+        { "manager-address", required_argument, NULL,
+                                                GETOPT_VAL_MANAGER_ADDRESS },
+        { "mtu",             required_argument, NULL, GETOPT_VAL_MTU },
+        { "help",            no_argument,       NULL, GETOPT_VAL_HELP },
+        { "plugin",          required_argument, NULL, GETOPT_VAL_PLUGIN },
+        { "plugin-opts",     required_argument, NULL, GETOPT_VAL_PLUGIN_OPTS },
+        { "password",        required_argument, NULL, GETOPT_VAL_PASSWORD },
+        { "key",             required_argument, NULL, GETOPT_VAL_KEY },
 #ifdef __linux__
-        { "mptcp",           no_argument,       0, 0 },
+        { "mptcp",           no_argument,       NULL, GETOPT_VAL_MPTCP },
 #endif
-        {                 0,                 0, 0, 0 }
+        { NULL,              0,                 NULL, 0 }
     };
 
     opterr = 0;
 
     USE_TTY();
 
-    while ((c = getopt_long(argc, argv, "f:s:p:l:k:t:m:b:c:i:d:a:n:huUvA6",
-                            long_options, &option_index)) != -1) {
+    while ((c = getopt_long(argc, argv, "f:s:p:l:k:t:m:b:c:i:d:a:n:huUv6A",
+                            long_options, NULL)) != -1) {
         switch (c) {
-        case 0:
-            if (option_index == 0) {
-                fast_open = 1;
-            } else if (option_index == 1) {
-                LOGI("initializing acl...");
-                acl = !init_acl(optarg);
-            } else if (option_index == 2) {
-                manager_address = optarg;
-            } else if (option_index == 3) {
-                mtu = atoi(optarg);
-                LOGI("set MTU to %d", mtu);
-            } else if (option_index == 4) {
-                usage();
-                exit(EXIT_SUCCESS);
-            } else if (option_index == 5) {
-                plugin = optarg;
-            } else if (option_index == 6) {
-                plugin_opts = optarg;
-            } else if (option_index == 7) {
-                mptcp = 1;
-                LOGI("enable multipath TCP");
-            }
+        case GETOPT_VAL_FAST_OPEN:
+            fast_open = 1;
+            break;
+        case GETOPT_VAL_ACL:
+            LOGI("initializing acl...");
+            acl = !init_acl(optarg);
+            break;
+        case GETOPT_VAL_MANAGER_ADDRESS:
+            manager_address = optarg;
+            break;
+        case GETOPT_VAL_MTU:
+            mtu = atoi(optarg);
+            LOGI("set MTU to %d", mtu);
+            break;
+        case GETOPT_VAL_PLUGIN:
+            plugin = optarg;
+            break;
+        case GETOPT_VAL_PLUGIN_OPTS:
+            plugin_opts = optarg;
+            break;
+        case GETOPT_VAL_MPTCP:
+            mptcp = 1;
+            LOGI("enable multipath TCP");
+            break;
+        case GETOPT_VAL_KEY:
+            key = optarg;
+            break;
+        case GETOPT_VAL_REUSE_PORT:
+            reuse_port = 1;
             break;
         case 's':
             if (server_num < MAX_REMOTE_NUM) {
@@ -1638,6 +1551,7 @@ main(int argc, char **argv)
         case 'p':
             server_port = optarg;
             break;
+        case GETOPT_VAL_PASSWORD:
         case 'k':
             password = optarg;
             break;
@@ -1679,15 +1593,16 @@ main(int argc, char **argv)
         case 'v':
             verbose = 1;
             break;
+        case GETOPT_VAL_HELP:
         case 'h':
             usage();
             exit(EXIT_SUCCESS);
-        case 'A':
-            auth = 1;
-            break;
         case '6':
             ipv6first = 1;
             break;
+        case 'A':
+            FATAL("One time auth has been deprecated. Try AEAD ciphers instead.");
+            break;
         case '?':
             // The option character is not recognized.
             LOGE("Unrecognized option: %s", optarg);
@@ -1720,6 +1635,9 @@ main(int argc, char **argv)
         if (password == NULL) {
             password = conf->password;
         }
+        if (key == NULL) {
+            key = conf->key;
+        }
         if (method == NULL) {
             method = conf->method;
         }
@@ -1735,9 +1653,6 @@ main(int argc, char **argv)
         if (plugin_opts == NULL) {
             plugin_opts = conf->plugin_opts;
         }
-        if (auth == 0) {
-            auth = conf->auth;
-        }
         if (mode == TCP_ONLY) {
             mode = conf->mode;
         }
@@ -1747,11 +1662,12 @@ main(int argc, char **argv)
         if (mptcp == 0) {
             mptcp = conf->mptcp;
         }
-#ifdef TCP_FASTOPEN
+        if (reuse_port == 0) {
+            reuse_port = conf->reuse_port;
+        }
         if (fast_open == 0) {
             fast_open = conf->fast_open;
         }
-#endif
 #ifdef HAVE_SETRLIMIT
         if (nofile == 0) {
             nofile = conf->nofile;
@@ -1766,14 +1682,17 @@ main(int argc, char **argv)
     }
 
     if (server_num == 0) {
-        server_host[server_num++] = NULL;
+        server_host[server_num++] = "0.0.0.0";
     }
 
-    if (server_num == 0 || server_port == NULL || password == NULL) {
+    if (server_num == 0 || server_port == NULL
+            || (password == NULL && key == NULL)) {
         usage();
         exit(EXIT_FAILURE);
     }
 
+    remote_port = server_port;
+
     if (plugin != NULL) {
         uint16_t port = get_local_port();
         if (port == 0) {
@@ -1823,10 +1742,6 @@ main(int argc, char **argv)
 #endif
     }
 
-    if (auth) {
-        LOGI("onetime authentication enabled");
-    }
-
     if (plugin != NULL) {
         LOGI("plugin \"%s\" enabled", plugin);
     }
@@ -1852,15 +1767,16 @@ main(int argc, char **argv)
 
     // setup keys
     LOGI("initializing ciphers... %s", method);
-    int m = enc_init(password, method);
+    crypto = crypto_init(password, key, method);
+    if (crypto == NULL)
+        FATAL("failed to initialize ciphers");
 
     // initialize ev loop
     struct ev_loop *loop = EV_DEFAULT;
 
     // setup udns
     if (nameserver_num == 0) {
-        nameservers[nameserver_num++] = "8.8.8.8";
-        resolv_init(loop, nameservers, nameserver_num, ipv6first);
+        resolv_init(loop, NULL, 0, ipv6first);
     } else {
         resolv_init(loop, nameservers, nameserver_num, ipv6first);
     }
@@ -1882,7 +1798,7 @@ main(int argc, char **argv)
         }
 
         int err = start_plugin(plugin, plugin_opts, server_str,
-                plugin_port, "127.0.0.1", server_port);
+                plugin_port, "127.0.0.1", server_port, MODE_SERVER);
         if (err) {
             FATAL("failed to start the plugin");
         }
@@ -1916,7 +1832,6 @@ main(int argc, char **argv)
             // Setup proxy context
             listen_ctx->timeout = atoi(timeout);
             listen_ctx->fd      = listenfd;
-            listen_ctx->method  = m;
             listen_ctx->iface   = iface;
             listen_ctx->loop    = loop;
 
@@ -1926,7 +1841,7 @@ main(int argc, char **argv)
             if (host && strcmp(host, ":") > 0)
                 LOGI("tcp server listening at [%s]:%s", host, server_port);
             else
-                LOGI("tcp server listening at %s:%s", host ? host : "*", server_port);
+                LOGI("tcp server listening at %s:%s", host ? host : "0.0.0.0", server_port);
 
             if (plugin != NULL) break;
         }
@@ -1940,12 +1855,11 @@ main(int argc, char **argv)
                 port = plugin_port;
             }
             // Setup UDP
-            init_udprelay(host, port, mtu, m,
-                    auth, atoi(timeout), iface);
+            init_udprelay(host, port, mtu, crypto, atoi(timeout), iface);
             if (host && strcmp(host, ":") > 0)
                 LOGI("udp server listening at [%s]:%s", host, port);
             else
-                LOGI("udp server listening at %s:%s", host ? host : "*", port);
+                LOGI("udp server listening at %s:%s", host ? host : "0.0.0.0", port);
         }
     }
 
